
from http import server
import os
import sys
import subprocess
import time
import traceback
import threading

try:
  import wx
except:
  traceback.print_exc()
  subprocess.run([
    sys.executable, *('-m pip install --user wxPython'.split())
  ])
  import wx
import wx.adv


import net_work

def maybe(callback):
  try:
    return callback()
  except:
    return traceback.format_exc()

def first_existing_file(*files):
  for f in files:
    if os.path.exists(f):
      return f
  return None

app = None
frame = None
server_t = None
server_t_exit_flag = False
server_work_req_folder = ""
server_work_return_folder = ""

class TaskBarGuiIcon(wx.adv.TaskBarIcon):
  
  def __init__(self, parent_frame):
    wx.adv.TaskBarIcon.__init__(self)

    self.parent_frame = parent_frame

    icon_file = first_existing_file(
      r'S:\Users\jmcateer\ToolResources\net_work_tray_icon.ico',
      r'C:\Temp\net_work_tray_icon.ico',
      '/tmp/net_work_tray_icon.ico',
    )
    if icon_file is not None:
      self.SetIcon(wx.Icon(icon_file, wx.BITMAP_TYPE_ICO))
    

  
  def CreatePopupMenu(self):
    global server_t_exit_flag, server_work_req_folder, server_work_return_folder
    m = wx.Menu()

    m.Append(wx.ID_NEW, "net-work server running", "net-work server running")
    m.AppendSeparator()

    m.Append(wx.ID_NEW, "Working tasks sent to {}".format(server_work_req_folder), "Working tasks sent to {}".format(server_work_req_folder))
    m.Append(wx.ID_NEW, "Returning tasks to {}".format(server_work_return_folder), "Returning tasks to {}".format(server_work_return_folder))
    m.AppendSeparator()

    m.Append(wx.ID_ANY, "Quit net-work server", "Quit net-work server")

    def on_quit_clicked(event):
      global server_t_exit_flag, frame

      # print('Exiting... (event={}, id={})'.format(event, event.GetId() ))
      
      try:
        if event.GetId() == wx.ID_NEW:
          return # Not an exit event
      except:
        traceback.print_exc()
      
      try:
        server_t_exit_flag = True
        time.sleep(3)
      except:
        traceback.print_exc()
      try:
        frame.onClose(event)
      except:
        traceback.print_exc()
      # sys.exit(0)

    self.Bind(wx.EVT_MENU, on_quit_clicked)

    return m

class DummyGuiFrame(wx.Frame):
  def __init__(self):
    wx.Frame.__init__(self, None, wx.ID_ANY, "", size=(1,1))
    panel = wx.Panel(self)
    self.tray_icon = TaskBarGuiIcon(self)
    self.Bind(wx.EVT_CLOSE, self.onClose)
  
  def onClose(self, evt):
    global server_t_exit_flag
    server_t_exit_flag = True
    self.tray_icon.RemoveIcon()
    self.tray_icon.Destroy()
    self.Destroy()


class ServerGuiApp(wx.App):
  def OnInit(self):
    global frame
    frame = DummyGuiFrame()
    self.SetTopWindow(frame)
    #frame.Center(wx.BOTH)
    #frame.Show(False)
    return True

def run_gui_server(args: list):
  global app, server_t

  app = ServerGuiApp(0)

  server_t = threading.Thread(target=run_server, args=(args, ))
  server_t.start()

  time.sleep(0.75) # Wait for run_server to possibly die b/c of bad arguments, if it does we also die.
  if not server_t.is_alive():
    return
  
  print('App constructed, running GUI loop...')
  app.MainLoop()

def run_server(args: list):
  global server_t_exit_flag, server_work_req_folder, server_work_return_folder
  # remove args until we hit the first one beginning with '-'
  while len(args) > 0 and not args[0].startswith('-'):
    args.pop(0)
  # Now turn args (which we assume to be like ['-req', '/path/to/a/dir', '-return', '/path/to/another/dir',])
  # into a dictionary
  args_d = {}
  for arg_flag, arg_value in zip(args[0::2], args[1::2]):
    arg_flag = arg_flag.replace('-', '') # --req and -req both turn into "req" in our argument dictionary.
    args_d[arg_flag] = arg_value

  work_req_dir = args_d['req']
  work_return_dir = work_req_dir
  if 'return' in args_d:
    work_return_dir = args_d['return']

  # TODO parse all other arguments

  # Save globals in case GUI is running
  server_work_req_folder = work_req_dir
  server_work_return_folder = work_return_dir

  nw = net_work.NetWork(work_req_dir, work_return_dir)
  # Poll for incoming work, run, return, etc.
  files_we_must_delete = []
  while not server_t_exit_flag:
    if len(files_we_must_delete) > 0:
      try:
        for file_name in files_we_must_delete:
          net_work.NetWork.safe_del(file_name)
        files_we_must_delete = [] # clear list
      except:
        traceback.print_exc()
    
    try:
      work_d = nw.poll_any_work_req_dicts(should_continue_polling_fn=lambda: not server_t_exit_flag)
      if work_d is None:
        continue # should_continue_polling_fn said to exit
      
      files_we_must_delete.append(
        nw.work_req_file(work_d['job_id']),
      )
      files_we_must_delete.append(
        nw.work_lock_file(work_d['job_id']),
      )

      if 'cmd' in work_d:
        cmd_args = work_d['cmd']
        proc = subprocess.run(cmd_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        nw.write_work_return_file(work_d['job_id'], {
          'returncode': proc.returncode,
          'stdout': maybe(lambda: proc.stdout.decode()),
          'stderr': maybe(lambda: proc.stderr.decode()),
        })
        
      else:
        raise Exception(f'Unknown work dictionary={work_d}')
    except:
      traceback.print_exc()
      time.sleep(1)
  print('Server thread exiting...')









import os
import sys
import subprocess
import random
import string
import json
import time
import traceback

class NetWork:
  '''
  Represents all the details needed to make remote calls and read remote responses.
  Provides both a syncronous and an asyncronous API.
  '''

  def gen_job_id(num_chars=8):
    return ''.join([random.choice(string.ascii_lowercase + string.digits) for _ in range(num_chars)])

  def read_json(json_f):
    with open(json_f, 'r') as fd:
      return json.load(fd)

  def safe_del(file_name):
    if os.path.exists(file_name):
      try:
        os.remove(file_name)
      except:
        traceback.print_exc()

  def __init__(self, shared_work_req_directory, shared_work_return_directory=None, work_req_prefix='call_', work_lock_prefix='lock_', work_return_prefix='return_', fatal_timeout_s=300):
    '''
    :param str shared_work_req_directory: Shared network directory where input requests and files will be written
    :param str shared_work_return_directory: Optional, defaults to shared_work_req_directory if unspecified. Directory where completed work will be written.
    :param str work_req_prefix: Prefix to use for input file names
    :param str work_return_prefix: Prefix to use for output file names
    '''
    self.shared_work_req_directory = shared_work_req_directory

    # If unspecified, out directory is the same as in directory.
    if shared_work_return_directory is None:
      self.shared_work_return_directory = shared_work_req_directory
    else:
      self.shared_work_return_directory = shared_work_return_directory

    if work_req_prefix.lower() == work_return_prefix.lower() or work_req_prefix.lower() == work_lock_prefix.lower():
      raise Exception('Refusing to use identical work_req_prefix, work_lock_prefix, work_return_prefix, please ensure these are different!')

    self.work_req_prefix = work_req_prefix
    self.work_lock_prefix = work_lock_prefix
    self.work_return_prefix = work_return_prefix
    self.fatal_timeout_s = fatal_timeout_s

    for req_d in [self.shared_work_req_directory, self.shared_work_return_directory]:
      if not os.path.exists(req_d):
        os.makedirs(req_d)

  def job_id_from_file_name(self, file_name):
    if self.shared_work_req_directory in file_name:
      file_name = file_name.replace(self.shared_work_req_directory, '').lstrip(os.sep)
    if self.work_req_prefix in file_name:
      file_name = file_name.replace(self.work_req_prefix, '')

    file_name = file_name.replace('.json', '')
    file_name = file_name.replace('.lock', '')

    return file_name

  def work_req_file(self, job_id):
    return os.path.join(
      self.shared_work_req_directory, self.work_req_prefix+job_id+'.json',
    )

  def work_lock_file(self, job_id):
    return os.path.join(
      self.shared_work_req_directory, self.work_lock_prefix+job_id+'.lock',
    )

  def work_return_file(self, job_id):
    return os.path.join(
      self.shared_work_return_directory, self.work_return_prefix+job_id+'.json',
    )
  
  def write_work_req_file(self, job_id, work_req_d):
    work_file = self.work_req_file(job_id)
    with open(work_file, 'w') as fd:
      json.dump(work_req_d, fd)

  def write_work_lock_file(self, job_id):
    if self.shared_work_req_directory in job_id:
      lock_f = job_id
    else:
      lock_f = self.work_lock_file(job_id)
    with open(lock_f, 'w') as fd:
      fd.write('LOCK')

  def write_work_return_file(self, job_id, work_return_d):
    work_file = self.work_return_file(job_id)
    with open(work_file, 'w') as fd:
      json.dump(work_return_d, fd)
    # Also unlock the work
    lock_f = self.work_lock_file(job_id)
    if os.path.exists(lock_f):
      os.remove(lock_f)

  def poll_work_return_file(self, job_id, poll_interval_s=0.25):
    if self.shared_work_return_directory in job_id:
      work_file = job_id # someone mis-passed data, but that's fine & detectable
    else:
      work_file = self.work_return_file(job_id)

    begin_s = int(time.time())
    while not os.path.exists(work_file):
      time.sleep(poll_interval_s)
      if int(time.time()) - begin_s > self.fatal_timeout_s:
        raise Exception('Timeout during poll_work_return_file!')

    with open(work_file, 'r') as fd:
      return json.load(fd)


  def run_cmd_sync(self, cmd_line: list):
    try:
      job_id = NetWork.gen_job_id()
      self.write_work_req_file(job_id, {
        'cmd': cmd_line,
      })
      work_return_d = self.poll_work_return_file(job_id)
      
      NetWork.safe_del( self.work_req_file(job_id) )
      NetWork.safe_del( self.work_lock_file(job_id) )
      NetWork.safe_del( self.work_return_file(job_id) )

      return work_return_d
    except:
      traceback.print_exc()
      return traceback.format_exc()
    return None

  def poll_any_work_req_dicts(self, poll_interval_median_s=0.5, poll_interval_dev=0.4, should_continue_polling_fn=None):
    poll_interval_min_s = max(0, poll_interval_median_s - poll_interval_dev)
    poll_interval_max_s = poll_interval_median_s + poll_interval_dev

    if should_continue_polling_fn is None:
      should_continue_polling_fn = lambda: True

    while should_continue_polling_fn():

      # We sleep for random times to reduce network lock-file race-cases.
      time.sleep(random.uniform(poll_interval_min_s, poll_interval_max_s))

      for file_name in os.listdir(self.shared_work_req_directory):
        if file_name.startswith(self.work_req_prefix):
          json_file_path = os.path.join(self.shared_work_req_directory, file_name)
          # Found one! Is this claimed by another processing node?
          job_id = self.job_id_from_file_name(file_name)
          lock_f = self.work_lock_file(job_id)
          if not os.path.exists(lock_f):
            # We're first, claim it!
            self.write_work_lock_file(lock_f)
            # Return the work dictionary
            work_d = NetWork.read_json(json_file_path)
            work_d['job_id'] = job_id
            return work_d




